
# add eunms
# add exceptions for send and recv
# add finally
# add GUI ;-)

### Global Definitions

RECV_SIZE = 1024

#######################

### Global Imports

import socket

#######################
class Cell:
    def __init__(self,x,y):
        self.x = x
        self.y = y
        self.state = 0 #empty
        self.mysub = 0
    def SetSub(self,sub):
        self.state = 1 #sub
        self.mysub = sub
    def Hit(self):
        if self.state == 0:
            self.state = 2 #miss
            return 0
        elif self.state == 1: #if cell has a not-hit-yet sub
            self.state = 3 #hit
            self.mysub.Hit()
            return 1
    def SetDead(self):
        self.state = 4 #dead
    def IsMe(self,c):
        return self == c
    def IsHit(self):
        return self.state == 3
    def IsAdjacent(self,c):
        return  ( abs(self.x - c.x) == 1 and (self.y == c.y) ) or ( abs(self.y - c.y) == 1 and (self.x == c.x) )
    def ToString(self,is_private=1):
        if self.state == 0: # empty
            return "."
        elif self.state == 1: # sub
            if is_private:
                return "0"
            else:
                return "."
        elif self.state == 2: # miss
            return "*"
        elif self.state == 3: # hit
            return "@"
        elif self.state == 4: # dead
            return "#"
    def SockString(self):
        buf = ""
        ## - Remember to write about this kind of initialization as python type declaration
        buf += str(self.x) + ":" + str(self.y)
        return buf    
        

class Sub:
    def __init__(self,cells):
        self.cells = cells
        self.is_dead = 0
    def Hit(self):
        hitcells = filter(lambda x: x.IsHit(), self.cells)
        if len(hitcells) == len(self.cells):
            self.is_dead = 1
    def SockString(self):
        buf = ",".join(map(lambda x: x.SockString(), self.cells))
        return buf
        
        
class Board:
    def __init__(self,columns,rows):
        self.columns = columns
        self.rows = rows
        self.cells = []
        for y in range(rows):
            self.cells.append([])
            for x in range(columns):
                self.cells[y].append(Cell(x,y))
        self.subs = []
        self.is_done = 0
    def AddSub(self,cells):
        # go over cells - check space legality
        for c in cells:
            # check that the cell is empty
            if c.state != 0: # not empty
                return 0,"Sub is located on a non empty cell!"
            # check that its neighbors are empty
            for x_ in range(c.x-1,c.x+2):
                for y_ in range(c.y-1,c.y+2):
                    if not (x_ == c.x and y_ == c.y) and not self.IsCellEmpty(x_,y_):
                        return 0,"Sub is located too close to other subs!"
            # check that all cells of the sub are adjacent
            found_adjacent = 0
            for c_ in cells:
                if c.IsAdjacent(c_):
                    found_adjacent = 1
            if not found_adjacent and len(cells) > 1:
                return 0,"cells do not construct a sub!"
        # all is legal - create a sub on the board
        sub = Sub(cells)
        self.subs.append(sub)
        # go over cells and mark submarine
        for c in cells:
            c.SetSub(sub)
        return 1,"OK"
    def IsCellEmpty(self,x,y):
        """ This method returns 1 if the cell is empty,
        or it is located outside the board's perimeters """
        if x < 0 or x >= self.columns or y < 0 or y >= self.rows: # out of range
            return 1
        elif self.cells[y][x].state == 0: # empty cell
            return 1
        return 0
    def Hit(self,x,y):
        """ This method returns 1 if hit was successfull (meaning extra turn) """
        c = self.cells[y][x]
        if c.Hit():
            if c.mysub.is_dead:
                print "S"
                # mark all cells of sub as Dead
                map(lambda x: x.SetDead(), c.mysub.cells)
                # mark all neighboring cells as Miss
                for c_ in c.mysub.cells:
                    for x_ in range(c_.x-1,c_.x+2): # note that range needs +2
                        for y_ in range(c_.y-1,c_.y+2):
                            self.MarkAsMiss(x_,y_)
                # check if all subs are dead
                if reduce( lambda x,y:x and y , map(lambda z:z.is_dead,self.subs) ):
                    self.is_done = 1
                return 2 # killed sub
            return 1 # hit sub
        return 0 # missed
    def MarkAsMiss(self,x,y):
        if x < 0 or x >= self.columns or y < 0 or y >= self.rows: # out of range
            pass
        elif self.cells[y][x].state == 0: # empty cell
            self.cells[y][x].state = 2 # miss
    def ToString(self, is_private=1):
        s = "  "
        s += "".join(map(lambda x: str(x % 10) ,range(self.columns)))
        s += "\n"
        for y in range(self.rows):
            s += str(y % 10) + " "
            for x in range(self.columns):
                s += self.cells[y][x].ToString(is_private)
            s += "\n"
        return s

class Player:
        def __init__(self,board_size=0,subs_level=0):
            self.sock = socket.socket(socket.AF_INET,socket.SOCK_STREAM)
            self.board_size = board_size
            self.subs_level = subs_level
        def ConnnectPlayer(self,ip,port):
            try:
                self.sock.connect((ip,port))
                buf = self.sock.recv(20)
            except socket.error, (errno, strerror):
                print "Network problems:",strerror
                return 0
            size_s, subs_level_s = buf.split(' ')
            self.board_size = int(size_s)
            self.subs_level = int(subs_level_s)
            # FF read num of subs and sizes using ConfigParser
            self.my_board = Board(self.board_size,self.board_size)
            self.other_board = Board(self.board_size,self.board_size)
            return 1
        def RecvSubs(self):
            buf = self.sock.recv(RECV_SIZE)
            for sub_str in buf.split(' '):
                cells = []
                for cell_str in sub_str.split(','):
                    x,y = cell_str.split(':')
                    cells.append(self.other_board.cells[int(y)][int(x)])
                # we assume that errors are detected at opponent's side
                self.other_board.AddSub(cells)
        def SendSubs(self):
            buf = " ".join(map(lambda x:x.SockString(), self.my_board.subs))
            self.sock.send(buf)
        def AcceptPlayer(self,port):
            try:
                self.sock.bind(('', port))
                self.sock.listen(1)
                self.sock, addr = self.sock.accept()
                buf = str(self.board_size) + " " + str(self.subs_level)
                self.sock.send(buf)
            except socket.error, (errno, strerror):
                print "Network problems:",strerror
                return 0
            self.my_board = Board(self.board_size,self.board_size)
            self.other_board = Board(self.board_size,self.board_size)
            # FF send num of subs
            return 1
        def CheckCoordinates(self,string):
            s_x,s_y = string.split(' ')
            x,y = int(s_x), int(s_y)
            if x >= self.board_size or x < 0 or y >= self.board_size or y < 0:
                raise ValueError, "Coordinates should be between 0 and " + str(self.board_size)
            return x,y
        def GetSubsFromPlayer(self):
            print "Please enter the locations of your subs:"
            for sub_size in range(self.subs_level,0,-1):
                for i in range(self.subs_level-sub_size+1):
                    rc=0
                    while (rc is 0):
                        print "Please enter sub of size",sub_size
                        sub = []
                        for c in range(sub_size):
                            x,y = my_input("Please enter coordinates of cell "+str(c)+": ",self.CheckCoordinates)
                            sub.append(self.my_board.cells[y][x])
                        rc, err = self.my_board.AddSub(sub)
                        if (rc is 0):
                            print "Bad input given:", err
                        else:
                            print "Your Board:"
                            print self.my_board.ToString()
        def GetMoveFromPlayer(self):
            x,y = my_input("Please enter your guess: ",self.CheckCoordinates)
            self.SendMove(x,y)
            hit = self.other_board.Hit(x,y)
            if hit == 1:
                print "You hit!"
            elif hit == 2:
                print "You hit and sunk a submarine!"
            else:
                print "You missed..."
            self.PrintBoards()
            return hit
        def GetMoveFromOtherPlayer(self):
            print "Waiting for opponent to move..."
            x,y = self.RecvMove()
            hit = self.my_board.Hit(int(x),int(y))
            if hit == 1:
                print "You were hit!"
            elif hit == 2:
                print "You were hit and lost a submarine!"
            else:
                print "Opponent missed..."
            self.PrintBoards()
            return hit
        def SendMove(self,x,y):
            self.sock.send(str(x)+":"+str(y))
        def RecvMove(self):
            buf = self.sock.recv(10)
            return buf.split(':')
        def PrintBoards(self):
            print "Your Board:"
            print self.my_board.ToString()
            print "Other's Board:" 
            print self.other_board.ToString(0)
        def PlayGame(self,other_player_start=0):
            # Till end of game
            while (1):
                if (other_player_start):
                    # Get move from other player
                    while ( self.GetMoveFromOtherPlayer() ):
                        if (self.my_board.is_done):
                            print "You lost!"
                            return
                        else:
                            print "Opponent has another turn"
                # Get move from player
                while ( self.GetMoveFromPlayer() ):
                    if (self.other_board.is_done):
                        print "You won!"
                        return
                    else:
                        print "You have another turn"
                other_player_start=1
        def RunServer(self,port):
            # wait for connection from a player
            print "Waiting for opponent..."
            if (not self.AcceptPlayer(port)):
                return
            # Get Subs from player
            print "Opponent connected..."
            self.GetSubsFromPlayer()
            # send subs to other player
            self.SendSubs()
            # read subs from other player
            print "Waiting for opponent to select subs..."
            self.RecvSubs()
            print "... Done."
            # play the game!
            self.PlayGame()
        def RunClient(self,ip,port):
            # connect to server
            print "Connecting to play..."
            if (not self.ConnnectPlayer(ip, port)):
                return
            print "... Done."
            # Get Subs from player
            self.GetSubsFromPlayer()
            # read subs from other player
            print "Waiting for player to select subs..."
            self.RecvSubs()
            print "... Done."
            # send subs to other player
            self.SendSubs()
            # play the game! (other player starts)
            self.PlayGame(1)

def my_input(message,type_func=str):
    ok = 0
    while not ok:
        try:
            x = raw_input(message)
            y = type_func(x)
            ok = 1
        except ValueError,e:
            print "Bad value given: ",e," Try again."
    return y

def int_input(message):
    return my_input(message,int)

def is_server_client(string):
    if not string.lower() in ["s","c"]:
        raise ValueError, "not 's' or 'c'."
    return string.lower()

######################################
#
# MAIN PROGRAM
#
######################################

try:
    s_c = my_input("Press S to run a server, C to run a client: ", is_server_client)
    if s_c == "s":
        port = int_input("Which port to listen on? ")
        board_size = int_input("What is the board size? ")
        subs_level = int_input("What is the submarines level? ")
        s = Player(board_size,subs_level)
        s.RunServer(port)
    else:
        ip = my_input("Which ip to connect to? ")
        port = int_input("Which port to connect to? ")
        s = Player()
        s.RunClient(ip,port)
except KeyboardInterrupt:
    print "\n\nProgram stopped.\nGoodBye."

            

        
